package sshwrapper

import (
	"context"
	"fmt"
	"github.com/it-novum/nsta/nstautil"
	"io"
	"log"
	"os"
	"os/user"
	"path"
	"sync"
	"sync/atomic"
	"syscall"
	"testing"
	"time"

	"github.com/gliderlabs/ssh"
)

type serverStr struct {
	mtx    sync.Mutex
	server *ssh.Server
	port   int
}

var sshServer serverStr

func testSSH() {
	var err error
	sshServer.mtx.Lock()
	defer sshServer.mtx.Unlock()

	if sshServer.server == nil {
		if sshServer.port, err = nstautil.DynamicPort(); err != nil {
			log.Fatalln(err)
		}
		log.Println("starting ssh server on port ", sshServer.port)

		forwardHandler := &ssh.ForwardedTCPHandler{}

		server := ssh.Server{
			LocalPortForwardingCallback: ssh.LocalPortForwardingCallback(func(ctx ssh.Context, dhost string, dport uint32) bool {
				log.Println("Accepted forward", dhost, dport)
				return true
			}),
			Addr: fmt.Sprintf("127.0.0.1:%d", sshServer.port),
			Handler: ssh.Handler(func(s ssh.Session) {
				_, _ = io.WriteString(s, "Remote forwarding available...\n")
			}),
			ReversePortForwardingCallback: ssh.ReversePortForwardingCallback(func(ctx ssh.Context, host string, port uint32) bool {
				log.Println("attempt to bind", host, port, "granted")
				return true
			}),
			PublicKeyHandler: func(ctx ssh.Context, key ssh.PublicKey) bool {
				log.Println("login granted")
				return true
			},
			RequestHandlers: map[string]ssh.RequestHandler{
				"tcpip-forward":        forwardHandler.HandleSSHRequest,
				"cancel-tcpip-forward": forwardHandler.HandleSSHRequest,
			},
		}

		go func() {
			log.Fatalln(server.ListenAndServe())
		}()

		timeout := time.NewTimer(time.Second * 2)
		for {
			select {
			case <-timeout.C:
				log.Fatalln("SSH Server Timeout")
			default:
				if nstautil.ConnectionTest("127.0.0.1", sshServer.port) {
					return
				}
			}
		}
	}
}

func TestSSHConn(t *testing.T) {
	testSSH()

	u, err := user.Current()
	if err != nil {
		t.Fatal(err)
	}
	pk := path.Join(u.HomeDir, ".ssh", "id_rsa")
	if _, err := os.Stat(pk); os.IsNotExist(err) {
		t.Fatal("private key file does not exist")
	}

	c := Conn{
		ConnOpts: ConnOpts{
			Hostname:       "127.0.0.1",
			Port:           sshServer.port,
			Login:          u.Name,
			PrivateKeyPath: pk,
			RemotePort:     4730,
		},
		ClientName: "test",
	}

	port, err := c.Start(context.Background())
	if err != nil {
		t.Fatal(err)
	}
	defer c.Stop()

loopb:
	for {
		select {
		case <-time.NewTimer(time.Second * 5).C:
			t.Fatal("SSH Timeout")
		case p := <-port:
			// success
			if p != 0 {
				break loopb
			}
		}
	}
}

func TestContext(t *testing.T) {
	testSSH()

	u, err := user.Current()
	if err != nil {
		t.Fatal(err)
	}
	pk := path.Join(u.HomeDir, ".ssh", "id_rsa")
	if _, err := os.Stat(pk); os.IsNotExist(err) {
		t.Fatal("private key file does not exist")
	}

	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	c := Conn{
		ConnOpts: ConnOpts{
			Hostname:       "127.0.0.1",
			Port:           sshServer.port,
			Login:          u.Name,
			PrivateKeyPath: pk,
			RemotePort:     4730,
		},
		ClientName: "test",
	}

	port, err := c.Start(ctx)
	if err != nil {
		t.Fatal(err)
	}

loopa:
	for {
		select {
		case <-time.NewTimer(time.Second * 5).C:
			t.Fatal("SSH Timeout")
		case p := <-port:
			// success
			if p != 0 {
				break loopa
			}
		}
	}

	cancel()

	timeout := time.NewTimer(time.Second * 10)
	defer timeout.Stop()

loopb:
	for {
		select {
		case <-timeout.C:
			t.Fatal("Timeout")
		default:
			if atomic.LoadInt64(&c.pid) == ProcessStopped {
				log.Println("successfully stopped")
				break loopb
			}
			time.Sleep(time.Millisecond * 10)
		}
	}
}

func TestReconnect(t *testing.T) {
	testSSH()

	u, err := user.Current()
	if err != nil {
		t.Fatal(err)
	}
	pk := path.Join(u.HomeDir, ".ssh", "id_rsa")
	if _, err := os.Stat(pk); os.IsNotExist(err) {
		t.Fatal("private key file does not exist")
	}

	c := Conn{
		ConnOpts: ConnOpts{
			Hostname:       "127.0.0.1",
			Port:           sshServer.port,
			Login:          u.Name,
			PrivateKeyPath: pk,
			RemotePort:     4730,
			Reconnect:      time.Millisecond * 50,
		},
		ClientName: "test",
	}

	port, err := c.Start(context.Background())
	if err != nil {
		t.Fatal(err)
	}
	defer c.Stop()

loopa:
	for {
		select {
		case <-time.NewTimer(time.Second * 5).C:
			t.Fatal("SSH Timeout")
		case p := <-port:
			// success
			if p != 0 {
				break loopa
			}
		}
	}

	// reconnect
	_ = syscall.Kill(int(atomic.LoadInt64(&c.pid)), syscall.SIGKILL)

	havePort0 := false
	for {
		select {
		case <-time.NewTimer(time.Second * 10).C:
			t.Fatal("Expected new port")
		case p := <-port:
			// success
			if p == 0 {
				havePort0 = true
			} else {
				if !havePort0 {
					t.Fatal("missing port 0 reset")
				}
				return
			}
		}
	}
}
